# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.interfaces.callback_with_config import CallbackWithConfig

__all__ = [
    'CallbackWithConfig',
]
